@@ -54,6 +54,10 @@
  */
 #define PROCESS_OUTSTANDING_TIMEOUT 7200
 
+/* The maximum allowed size of a report file, if not overridden
+ */
+#define MAX_REPORT_FILESIZE 1000000000
+
 /* If true, we have an active Internet connection. True by default in case we
  * can't bring up GNetworkMonitor */
 static gboolean online_state = TRUE;
@@ -198,6 +202,9 @@ static const char* unacceptable_fields[] = {
 
     "Title",
 
+    /* Redundant with the addition of pa-info LP: #1893899 */
+    "AlsaInfo",
+
     NULL
 };
 
@@ -241,8 +248,12 @@ size_t
 server_response (char* ptr, size_t size, size_t nmemb, void* s)
 {
     GString *resp = (GString *) s;
-    g_string_append (resp, ptr);
-    return size * nmemb;
+    if (size != 1) {
+        g_warning("CURLOPT_WRITEFUNCTION callback received size !=1, against spec!\n");
+        return 0;
+    }
+    g_string_append_len (resp, ptr, nmemb);
+    return nmemb;
 }
 
 void
@@ -343,14 +354,21 @@ upload_report (const char* message_data, uint32_t message_len, GString *s)
         return response_code;
 }
 
-void
-destroy_key_and_value (gpointer key, gpointer value, gpointer user_data)
+gsize
+get_report_max_size (void)
 {
-    if (key)
-        g_free (key);
-    /* The value may be "", which is allocated on the stack. */
-    if (value && *(char*)value != '\0')
-        g_free (value);
+    const char* value = NULL;
+    gsize max_size;
+
+    value = g_getenv ("REPORT_MAX_SIZE");
+    if (value == NULL)
+        return MAX_REPORT_FILESIZE;
+
+    max_size = atol(value);
+    if (max_size == 0)
+        return MAX_REPORT_FILESIZE;
+
+    return max_size;
 }
 
 GHashTable*
@@ -371,22 +389,48 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
     gchar* token_p = NULL;
     char* key = NULL;
     char* value = NULL;
+    char* old_value = NULL;
     gchar* value_p = NULL;
     GError* err = NULL;
     gchar* end = NULL;
     size_t value_length;
     size_t value_pos;
+    struct stat st;
+    int fd;
+    int res;
+    gsize max_size;
 
     g_return_val_if_fail (report_path, NULL);
 
-    if (g_file_test (report_path, G_FILE_TEST_IS_SYMLINK) ||
-        !g_file_test (report_path, G_FILE_TEST_IS_REGULAR)) {
+    max_size = get_report_max_size();
+
+    fd = open(report_path, O_RDONLY | O_NOFOLLOW);
+    res = fstat(fd, &st);
+    if (res == -1) {
         g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
-                     "%s is a symlink or is not a regular file.", report_path);
+                     "%s could not be opened.", report_path);
+        if (fd >= 0)
+            close(fd);
         return NULL;
     }
+
+    if(!S_ISREG(st.st_mode)) {
+        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
+                     "%s is not a regular file.", report_path);
+        close(fd);
+        return NULL;
+    }
+
+    /* Limit the size of a report */
+    if(st.st_size > max_size) {
+        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
+                     "%s is too big to be processed.", report_path);
+        close(fd);
+        return NULL;
+    }
+
     /* TODO handle the file being modified underneath us. */
-    fp = g_mapped_file_new (report_path, FALSE, &err);
+    fp = g_mapped_file_new_from_fd (fd, FALSE, &err);
     if (err) {
         g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                      "Unable to map report: %s", err->message);
@@ -396,8 +440,17 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
 
     contents = g_mapped_file_get_contents (fp);
     file_len = g_mapped_file_get_length (fp);
+
+    /* Check report size again to make sure */
+    if(file_len > max_size) {
+        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
+                     "%s is too big to be processed.", report_path);
+        goto error;
+    }
+
     end = contents + file_len;
-    hash_table = g_hash_table_new (g_str_hash, g_str_equal);
+    hash_table = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                        g_free, g_free);
     p = contents;
 
     while (p < end) {
@@ -432,11 +485,17 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
                                  0, "Report value too long.");
                     goto error;
                 }
+                g_hash_table_steal (hash_table, key);
                 value = g_realloc (value, value_pos + 1 + value_length + 1);
                 value_p = value + value_pos;
                 *value_p = '\n';
                 value_p++;
             } else {
+                /* Make sure we properly free the old empty string value */
+                old_value = g_hash_table_lookup (hash_table, key);
+                if (old_value)
+                    g_free (old_value);
+                g_hash_table_steal (hash_table, key);
                 value = g_realloc (value, value_length + 1);
                 value_p = value;
             }
@@ -447,7 +506,13 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
                 if (*c != '\t' && *c >= '\0' && *c < ' ')
                     *c = '?';
             value_p += value_length;
-            g_hash_table_insert (hash_table, key, value ? value : "");
+            if (g_hash_table_contains (hash_table, key) == TRUE) {
+                g_set_error (error,
+                    g_quark_from_static_string ("whoopsie-quark"),
+                    0, "Report key must not be a duplicate.");
+                goto error;
+            }
+            g_hash_table_insert (hash_table, key, value ? value : g_strdup(""));
             p = token_p + 1;
         } else {
             /* Reset the value pointer. */
@@ -517,10 +582,17 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
             }
             p = token_p + 1;
 
-            g_hash_table_insert (hash_table, key, value ? value : "");
+            if (g_hash_table_contains (hash_table, key) == TRUE) {
+                g_set_error (error,
+                    g_quark_from_static_string ("whoopsie-quark"),
+                    0, "Report key must not be a duplicate.");
+                goto error;
+            }
+            g_hash_table_insert (hash_table, key, value ? value : g_strdup(""));
         }
     }
     g_mapped_file_unref (fp);
+    close(fd);
 
     /* Remove entries that we don't want to send */
     if (!full_report) {
@@ -538,8 +610,7 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
                     continue;
             }
 
-            g_hash_table_iter_steal(&iter);
-            destroy_key_and_value(key, value, NULL);
+            g_hash_table_iter_remove(&iter);
         }
     }
 
@@ -547,10 +618,10 @@ parse_report (const char* report_path, gboolean full_report, GError** error)
 
 error:
     if (hash_table) {
-        g_hash_table_foreach (hash_table, destroy_key_and_value, NULL);
         g_hash_table_destroy (hash_table);
     }
     g_mapped_file_unref (fp);
+    close(fd);
     return NULL;
 }
 
@@ -718,7 +789,6 @@ parse_and_upload_report (const char* crash_file)
         g_string_free (s, TRUE);
     }
 
-    g_hash_table_foreach (report, destroy_key_and_value, NULL);
     g_hash_table_destroy (report);
 
     return success;
